% MonteCarloVerification.m
%
% This file generates the Monte Carlo results in Figure 4 shown in the following submission:
%
% B. Hu and P. Seiler, "Pivotal Decomposition in Reliability
% Analysis of Fault Tolerant Control Systems", submitted to IEEE 
% transactions on Reliability, 2014.
%
%There are totally 6 testing points. To run this file, one needs to pick up
%one testing point. The 1st, 5th, 6th testing points take roughly 20~25
%mins to run on a standard desktop. The 2nd and 4th testing points have
%smaller probability, more replications and take roughly 4 hours to run.
%The 3rd testing point has smallest probability, most replications and
%takes roughly 40 hours to run.
%

tic;
frequency=50;           % Samle frequency, Hz
dt = 1/frequency;       % Sample time, sec
N = 3600/dt;            % Number of samples per hour
N0 = 10;                % Maximum Detection Delay
n=6;                    % component number
% Per-frame actuator failure rate (geometric failure distribution)
MTBF = 3.6e6/2.5;           % Mean-time between failure, sec
q = 1-exp(-dt/MTBF); 

% different testing points on the normalized threshold
HMC=[2.7 5.2 8 10.8 13.2 18];
% different replications required for different testing points
N_replication=[50000 500000 5000000 500000 50000 50000];
% pick up one testing point by setting i=1 or 2 or 3 or 4 or 5 or 6
i=6;
% get the testing point data
H=HMC(i);
N_r=N_replication(i);

f=10;
clear sys
for i=1:N_r
  %generate the component failure time  
x=random('geo',q,1,6)+1;
id=(x>N);


%the number of actuators which are working during the N-step window
num_Working=length(find(id==1));
%every actuator is working then the system is working
if min(id(:))==1
    sys(i)=0;
    % more than 1 actuator fails then the system fails 
elseif num_Working<=4
    sys(i)=1;
else 
    % when 1 actuator fails, the analysis is more involved
    % t is the switch time of the FTC system
    t=0;
    sys(i)=0;
    % find the failed actuator
   Failure_Actuator=find(id==0);
    % the mode denotes the controller the system is using
   mode=0;
   % define a counter to track whether the FTC system switches its controller or not 
     counter_switch=0;
     clear Act_State residual;
     % generate the actuator health states based on the failure time
  for j=1:6
    if id(j)==1
Act_State(j,:)=zeros(1,N);
else
    Act_State(j,:)=[zeros(1,x(j)-1) zeros(1,N-x(j)+1)+1];
    end
  end
  %generate the FDI residuals based on the component health states
  residual=Act_State(3:6,:)*0;
  for j=1:N
      %add fault signal if the actuator fails
              residual(:,j)=randn(4,1)+f*Act_State(3:6,j);
     % switch the controller at the first detected component failure
  if counter_switch==0 & max(abs(residual(:,j)))>H
  counter_switch=1;
  mode=min(find(abs(residual(:,j))>H))+2;
  %denote the switch time
  t=j;
  end
  % switch to the wrong controller then the system fails
  if mode~=Failure_Actuator &mode~=0
  sys(i)=1;continue;
  % switch later than the maximum detection delay then the system fails
  elseif mode==Failure_Actuator & t>=x(Failure_Actuator)+N0
      sys(i)=1;continue;
  end
  
  end
  % no switch in case of certain component failures, the system fails
  if mode==0 & Failure_Actuator>2
  sys(i)=1;
  end
  
  
  
end


end
%output the system failure probability
mean(sys(:))
toc;

